const WasmCRDTAsync = import("drawing-crdt")

import { spreadErasureIntervals } from "./erasure.js"
import P2PMesh from "./p2p-mesh.js"

Array.prototype.remove = function(elem) {
  const index = this.indexOf(elem)

  if (index > -1) {
    return this.splice(index, 1)[0]
  }

  return undefined
}

export default class WasmCRDTWrapper {
  constructor(WasmCRDT, room, interval) {
    this.room = room
    this.mesh = null

    this.users = {
      synced: [],
      syncing: null,
      check: false,
      waiting: [],
    }

    this.crdt = new WasmCRDT({
      on_stroke: (stroke_id, points) => {
        stroke_id = (" " + stroke_id).slice(1)

        const detail = { id: stroke_id, points }

        this.room.dispatchEvent(new CustomEvent("addOrUpdatePath", { detail }))
      },
      on_interval: (stroke_id, intervals) => {
        const detail = {
          id: stroke_id,
          intervals: spreadErasureIntervals(intervals),
        }

        this.room.dispatchEvent(
          new CustomEvent("removedIntervalsChange", {
            detail,
          }),
        )
      },
      on_deltas: (deltas) => this.mesh.broadcast(deltas, true),
      on_deltas_from_state: (uid, deltas) => {
        this.mesh.send(
          uid,
          {
            type: "sync step 2",
            message: deltas,
          },
          true,
        )

        this.room.dispatchEvent(
          new CustomEvent("weSyncedWithPeer", { detail: uid }),
        )
      },
    })

    this.interval = setInterval(() => {
      this.crdt.fetch_events()
      this.crdt.fetch_deltas()
    }, interval)
  }

  destroy() {
    clearInterval(this.interval)
    this.interval = null

    this.mesh.disconnect()
    this.mesh = null

    this.crdt.free()
    this.crdt = null

    this.room = null
  }

  static async initialise(room, options) {
    const { WasmCRDT } = await WasmCRDTAsync

    room.crdt = new WasmCRDTWrapper(
      WasmCRDT,
      room,
      (options.wasm && options.wasm.interval) || 0,
    )
    room.crdt.mesh = new P2PMesh(room.crdt, options)
  }

  getUserID() {
    return this.crdt.get_user()
  }

  setUserID(uid) {
    const success = this.crdt.set_user(uid)

    if (success) {
      this.room.dispatchEvent(new CustomEvent("allocateOwnID", { detail: uid }))
    }

    return success
  }

  fetchDrawingEvents() {
    return this.crdt.fetch_events()
  }

  addPath([x, y, w, colour]) {
    return this.crdt.add_stroke(x, y, w, colour)
  }

  extendPath(pathID, [x, y, w, colour]) {
    return this.crdt.add_point(pathID, x, y, w, colour)
  }

  endPath(pathID) {
    this.crdt.end_stroke(pathID)
  }

  extendErasureIntervals(pathID, newIntervals) {
    newIntervals.forEach(([from, to]) =>
      this.crdt.erase_stroke(pathID, from, to),
    )
  }

  getPathIDs() {
    return this.crdt.get_stroke_ids()
  }

  getPathPoints(pathID) {
    return this.crdt.get_stroke_points(pathID)
  }

  getErasureIntervals(pathID) {
    return this.crdt.get_stroke_intervals(pathID)
  }

  userJoined(uid) {
    if (
      this.users.syncing == uid ||
      this.users.synced.includes(uid) ||
      this.users.waiting.includes(uid)
    ) {
      return
    }

    this.users.waiting.push(uid)

    this._continueSync()

    this.room.dispatchEvent(new CustomEvent("userJoin", { detail: uid }))
  }

  userLeft(uid) {
    if (this.users.syncing == uid) {
      this.users.syncing = null
      this.users.check = false

      this._continueSync()
    } else if (
      !this.users.synced.remove(uid) &&
      !this.users.waiting.remove(uid)
    ) {
      return
    }

    this.room.dispatchEvent(new CustomEvent("userLeave", { detail: uid }))
  }

  receiveMessage(uid, message) {
    if (message instanceof Uint8Array) {
      return this.crdt.apply_deltas(message)
    }

    const { type, message: _message } = message

    if (type == "sync step 1" && _message instanceof Uint8Array) {
      this.crdt.fetch_deltas_from_state_vector(uid, _message)
    } else if (type == "sync step 2" && _message instanceof Uint8Array) {
      this.users.check = this.crdt.apply_deltas(_message)

      if (this.users.syncing == uid) {
        this.users.syncing = null

        if (this.users.check) {
          this.users.synced.forEach((user) =>
            this.mesh.send(user, { type: "sync check" }, false),
          )
          this.users.waiting.forEach((user) =>
            this.mesh.send(user, { type: "sync check" }, false),
          )
        }

        this.users.check = false

        this.users.synced.push(uid)

        this._continueSync()

        this.room.dispatchEvent(
          new CustomEvent("peerSyncedWithUs", { detail: uid }),
        )
      }
    } else if (type == "sync check") {
      this.users.synced.remove(uid)
      this.users.waiting.remove(uid)

      this.users.waiting.unshift(uid)

      if (this.users.syncing == uid) {
        this.users.syncing = null
        this.users.check = false

        this._continueSync()
      }
    }
  }

  _continueSync() {
    if (this.users.syncing != null || this.users.waiting.length <= 0) {
      return
    }

    this.users.syncing = this.users.waiting.shift()
    this.users.check = false

    this.mesh.send(
      this.users.syncing,
      {
        type: "sync step 1",
        message: this.crdt.get_state_vector(),
      },
      false,
    )

    this.room.dispatchEvent(
      new CustomEvent("waitingForSyncStep", { detail: this.users.syncing }),
    )
  }

  reportConnectionQuality(uid, quality) {
    this.room.dispatchEvent(
      new CustomEvent("userConnection", { detail: { id: uid, quality } }),
    )
  }
}

export const benchmark = {
  blocksize: Number.MAX_SAFE_INTEGER,
  eventsGC: Number.MAX_SAFE_INTEGER,
  syncStep1: Uint8Array.of(
    133,
    164,
    117,
    117,
    105,
    100,
    217,
    36,
    55,
    101,
    98,
    98,
    55,
    53,
    54,
    51,
    45,
    50,
    54,
    100,
    55,
    45,
    52,
    97,
    100,
    101,
    45,
    56,
    56,
    54,
    57,
    45,
    52,
    52,
    102,
    102,
    101,
    100,
    57,
    49,
    102,
    99,
    55,
    54,
    167,
    109,
    101,
    115,
    115,
    97,
    103,
    101,
    196,
    31,
    130,
    164,
    116,
    121,
    112,
    101,
    171,
    115,
    121,
    110,
    99,
    32,
    115,
    116,
    101,
    112,
    32,
    49,
    167,
    109,
    101,
    115,
    115,
    97,
    103,
    101,
    196,
    3,
    0,
    0,
    0,
    165,
    115,
    108,
    105,
    99,
    101,
    0,
    166,
    108,
    101,
    110,
    103,
    116,
    104,
    31,
    170,
    99,
    111,
    109,
    112,
    114,
    101,
    115,
    115,
    101,
    100,
    194,
  ),
}
